%{
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h> 
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>     
#include <ctype.h>
#include <mix.h>     
#include <mixasm.h>

#define update_memory(A) \
             memory[location_counter]=(A); lines[lineno]=location_counter++; 
char  *progname;
char  *source_file = NULL;
#define LIST_SUFFIX ".lst"
char  *listfile = "mixal.lst";
FILE  *lst;

int lineno = 1;
int location_counter = 0;
unsigned word;
struct Symbol *s;
struct ref *p;
char literal[13];
char opcode[6];
unsigned  memory[MEMSIZE];
int  lines[MEMSIZE];
extern Symbol *symlist;

int list = 0;         /* Produce listing */
int xref = 0;         /* Produce cross-reference listing */
int writeraw = 0;     /* Write raw code */

FILE *create_temp(char *tmpl);
int eval_op(char *c);
int word_s(unsigned w);
void symbols(void);
void end_routine(int start);
int num(char *text);
int find_backward(char *name);
void find_forward(Symbol *f);
int eval_f(char *c);
void warning(char *s, char *t);
void listing();
void dumpsymtab(FILE *fp);
 
unsigned add_mix(), sub_mix(), div_mix(), mul_mix(), w_eval();

char alf[] =
     " ABCDEFGHInJKLMNOPQRepSTUVWXYZ0123456789.,()+-*/=$<>@;:'abcdfghi";
 
%}

%union {
     unsigned val;     /* WATCH OUT FOR THIS */
     struct Symbol *symp;
     char *txt;
}
%token <symp> DEFINED_SYMBOL
%token <symp> LOCAL_SYMBOL
%token <symp> FUTURE_REFERENCE
%token <symp> LOC    
%token <symp> END    
%token <val> UNDEF 
%token <val> DEF 
%token <val> NUMBER
%token <val> ASTERISK
%token <txt> TEXT
%token <txt> OPCODE
%token <txt> PSEUDO
%token <txt> ALF      
%token BOGUS
%left '-' '+'
%left '*' '/'
%left ':' SLASHES 
%nonassoc UMINUS

%type  <val>  expression 
%type  <val>  atomic_expression 
%type  <val>  a_part i_part f_part w_value
%type <symp>  literal_constant
%type <symp>  label              
%type  <txt>  op pseudo
%%

statement_list: statement comment '\n'           { lineno++; }
              | statement_list  statement comment '\n'  { lineno++; }
              ;

statement     : label op a_part i_part f_part
                {
		     word = eval_op(opcode);
		     word |= SIGN($3) | MAG($3)*B3 | $4*B2 | $5*B1;  
		     update_memory(word);
		}
              | label pseudo w_value
                {
		     if (!strcmp(opcode,"EQU")) {
			  /* if no label wizard debug */
			  if ($1 == NULL)
			       dumpsymtab(stderr);
			  else
			       $1->value = $3;
		    } else if (!strcmp(opcode,"ORIG")) {
			 location_counter = word_s($3);
		    } else if (!strcmp(opcode,"CON")) {
			 update_memory($3);
		    } else if (!strcmp(opcode,"END")) {
			 if ($1 != NULL)
			      $1->type = END; 
			 symbols();
			 end_routine(word_s($3));
		    }
		}                    
              | label ALF
                {
		     update_memory(num($2));
		}
              | ASTERISK  { ; }
              ;

label         : /* none */
                {
		     $$ = NULL;
		}
              | LOC
                {
		     if (localref($1)) {
			  yyerror("invalid label");
			  YYERROR;
		     } 
		     $$ = $1; 
		     $1->type = DEF; 
		     $1->line = lineno;
		     $1->value = location_counter;
		}
              ;

comment       :  /* nothing */
              | TEXT { ; } 
              ;

op            : OPCODE
                {
		     strcpy(opcode, $1);
		}
              ;

pseudo        : PSEUDO
                {
		     strcpy(opcode, $1);
		}
              ;
  
expression    : atomic_expression  
              | '+'  atomic_expression %prec UMINUS
                {
		     $$ = $2;
		}
              | '-'  atomic_expression %prec UMINUS
                {
		     $$ = $2 ^ MINUS;
		} 
              | expression '+' atomic_expression
                {
		     $$ = add_mix($1, $3);
		}
              | expression '-' atomic_expression
                {
		     $$ = sub_mix($1, $3);
		}
              | expression ASTERISK atomic_expression
                {
		     $$ = mul_mix($1, $3);
		}
              | expression '/' atomic_expression
                {
		     $$ = div_mix($1, $3);
		} 
              | expression SLASHES atomic_expression
                {
		     $$ = mul_mix($1, B5 / MAG($3));
		}
              | expression ':' atomic_expression
                {
		     $$ = add_mix(mul_mix($1, 8), $3);
		}
              ;

atomic_expression: NUMBER
              |    ASTERISK
                {
		     $$ = location_counter;
		}
              | DEFINED_SYMBOL
                {
		     $$ = $1->value;
		     if (xref)
			  $1->xref = addref($1->xref, location_counter,
					    lineno);
		}
              | LOCAL_SYMBOL
                {
		     $$ = $1->value = find_backward($1->name);
		}
              ;  

a_part        : /* empty */
                {
		     $$ = 0;
		}
              | expression
              | FUTURE_REFERENCE
                {
		     $1->reflist = addref($1->reflist,location_counter,lineno);
		     if (xref)
			  $1->xref = addref($1->xref, location_counter,
					    lineno);
		}
              | literal_constant
                { ; }
              ;    

i_part        : /* empty */
                {
		     $$ = 0;
		}
              | ',' expression
                {
		     $$ = $2;
		}
              ;    

f_part        : /* empty */
                {
		     $$ = eval_f(opcode);
		}
              | '('  expression ')'
                {
		     $$ = $2;
		}
              ;    

w_value       : expression f_part
                {
		     $$ = w_eval(0, $1, $2);
		}
              | w_value ',' expression f_part
                {
		     $$ = w_eval($1, $3, $4);
		}
              ;

literal_constant: '=' w_value '='
                {
		     sprintf(literal,"=%d=",word_s($2));
		     if (!(s = lookup(literal)))
			  s = install(literal,$2,lineno);
		     s->reflist = addref(s->reflist,location_counter,lineno);
		     if (xref)
			  s->xref = addref(s->xref, location_counter, lineno);
   		}
                         
              ;
    
%%

unsigned
add_mix(unsigned a, unsigned b)
{ 
     unsigned s, t;
     
     s = SIGN(a);
     t = SIGN(b);
     if (s != t) {
	  if (MAG(a) > MAG(b)) {
	       a = MAG(a) - MAG(b);   
	       t = s;
	  } else
	       a = MAG(b) - MAG(a);   
     } else 
	  a = MAG(a) + MAG(b);
     if (a > B5-1)
	  yyerror("Arithmetic result greater than 1073741823");
     return a|t;
}

unsigned 
sub_mix(unsigned a, unsigned b)
{
     b ^= MINUS;
     return add_mix(a, b);
}


unsigned
mul_mix(unsigned a, unsigned b)
{
     unsigned s;
     long long w;
     s = SIGN(a) == SIGN(b) ? PLUS : MINUS;
     if ((w = MAG(a)*MAG(b)) > B5-1)
	  yyerror("Arithmetic result greater than 1073741823");
     return s|MAG(w);
}
  
unsigned
div_mix(unsigned a, unsigned b)
{
     unsigned s;
     s = SIGN(a);
     return s|MAG(MAG(a)/MAG(b));
}
   
unsigned
w_eval(unsigned x, unsigned y,int f)
{
     int l, r;
     unsigned sign, mask;

     l = f/8;
     r = f%8;
     mask = B5 - 1;
     if (l < 0 || l > r || r > 5)
	  yyerror("Invalid F-part");
     if ( l == 0 ) {   /* Is sign included ? */
	  sign = MINUS & y;  
	  l = 1;
     } else sign = MINUS & x;       /* keep previous sign */
     y <<= (4 - r + l) * BYTESIZE;  /* take correct no of bytes from right
				       end */
     y = MAG(y);                    /* truncate to 5 bytes */
     y >>= (l - 1) * BYTESIZE;      /* position within word */
     mask <<= (4 - r + l) * BYTESIZE; /* do same for mask */
     mask = MAG(mask);                /* truncate to 5 bytes */
     mask >>= (l - 1) * BYTESIZE;  
     x &= ~mask;    /* mask off where y is to be inserted */
     x = MAG(x);                     /* discard sign */
     x |= y;        /* insert y */
     x |= sign;     /* insert sign */
     return x;
}

int
word_s(unsigned w)  /* convert mix word into a signed integer */
{
     if (w & MINUS)  w = -MAG(w);         
     return w;
}

int
yyerror(char *s)  /* called for yacc syntax error */
{
     warning(s, (char *) 0);
}

void
warning(char *s, char *t)
{
     fprintf(stderr, "%s: %s", progname, s);
     if (t)
	  fprintf(stderr, " %s", t);
     fprintf(stderr, " near line %d\n", lineno);
}

int
main(int argc, char *argv[])
{
     int lexdebug = 0;
     int i;
     char *p;
     
     progname = argv[0];
     
     for (i = 0; i < MEMSIZE; i++)
	  lines[i] = -1;
     while ((i = getopt(argc, argv, "d:hlrVx")) != EOF) {
	  switch (i) {
	  case 'd':
	       for (p = optarg; *p; p++) {
		    switch (*p) {
		    case 'y':
			 yydebug = 1;
			 break;

		    case 'l':
			 lexdebug = 1;
			 break;

		    default:
			 fprintf(stderr, "invalid debugging level: %c\n", *p);
			 break;
		    }
	       }
	       break;
	  case 'l':  /* produce a listing */
	       list = 1;
	       break;
	  case 'r':
	       writeraw = 1;
	       break;
	  case 'V':  /* show version */
	       printf("%s (%s)\n", progname, PACKAGE_STRING);	  
	       exit(0);
	  case 'x':  /* produce cross-reference listing */
	       xref = 1;
	       break;
	  case 'h':
	       printf("usage: %s [-d L][-lhrxV] [filename]\n", progname);
	       exit(0);
	  default:
	       fprintf(stderr, "usage: %s [-d L][-lhrxV] [filename]\n", progname);
	       exit(1);
	  }
     }
     if (argc > optind) {
	  FILE *fp;
	  source_file = argv[optind];
	  if ((fp = fopen(source_file, "r")) == NULL) {
	       fprintf(stderr,"can't open %s: %s\n", argv[1], strerror(errno));
	       exit(1);
	  }
	  init_lex(fp, lexdebug);
	  if (list) {
	       unsigned len;
	       char *p = strrchr(source_file, '.');
	       if (p)
		    len = p - source_file;
	       else
		    len = strlen(source_file);
	       listfile = malloc(len + sizeof(LIST_SUFFIX));
	       if (!listfile)
		    abort();
	       memcpy(listfile, source_file, len);
	       strcat(listfile, LIST_SUFFIX);
	  }
     } else
	  init_lex(stdin, lexdebug);
     if (list)
	  lst = create_temp("lstXXXXXX");
     yyparse();
     sort_symbols();
     if (list)
	  listing();
     else if (xref)
	  dumpsymtab(stderr);
}

FILE *
create_temp(char *tmpl)
{
     char *buf;
     char *tmp;
     int fd;
     FILE *fp;
    
     if ((tmp = getenv("TMP")) == NULL && (tmp = getenv("TEMP")) == NULL)
	  tmp = "/tmp";
     buf = malloc(strlen(tmp) + 1 + strlen(tmpl) + 1);
     if (!buf)
	  abort();
     sprintf(buf, "%s/%s", tmp, tmpl);

#ifdef HAVE_MKSTEMP
     {
	  int save_mask = umask (077);
	  fd = mkstemp (buf);
	  umask (save_mask);
     }
 #else
     if (mktemp (buf))
	  fd = open (filename, O_CREAT|O_EXCL|O_RDWR, 0600);
     else
	  fd = -1;
 #endif

     if (fd == -1) {
	  fprintf(stderr, "can't create temporary file %s: %s\n",
		  tmp, strerror(errno));
	  exit(1);
     }
     
     unlink(buf);
     free(buf);
     return fdopen(fd, "w+");
}

void
listing()
{
     FILE *fp;
     char buf[512];
     int i, l; 
     
     rewind(lst);
     
     fp = fopen(listfile, "w");
     if (fp == NULL) {
	  fprintf(stderr, "can't open %s: %s\n", listfile, strerror(errno));
	  exit(1);
     }
     
     for (l = 1; l <= lineno; l++) {
	  if ((i = lines[l]) != -1) {
	       int Av, Iv, Fv, Cv, Sv;
	       Sv = (memory[i] & MINUS) ? '-' : '+';
	       Av = (memory[i] & (B5 - 1))>>3*BYTESIZE;
	       Iv = (memory[i] & (B3 - 1))>>2*BYTESIZE;
	       Fv = (memory[i] & (B2 - 1))>>BYTESIZE;
	       Cv = (memory[i] & (B1 - 1));
	       fprintf(fp, "%4d: %c %4d %2d %2d %2d   %3d ",
		       i, Sv, Av, Iv, Fv, Cv, l);
	  } else
	       fprintf(fp, "%23s %3d ", " ", l);
	
	  if (fgets(buf, sizeof(buf), lst)) 
	       fputs(buf, fp);
	  else
	    fputs("\n", fp);
     }
     fputs("\n", fp);
     dumpsymtab(fp);
     fclose(fp);
}


void
dumpsymtab(FILE *fp)
{
     Symbol *s;
     ref *refp;
  
     fprintf(fp, "\nSymbol     Value  Line");
     if (xref) {
	  fprintf(fp, "  Referenced on line");
     }
     fprintf(fp, "\n");
     for (s = symlist; s; s = s->next) {
	  if (localref(s)) /* don't dump local symbol backward and forward
			      refs */
	       continue;

	  fprintf(fp, "%-10s  %4d %4d",
		  s->name, word_s(s->value), s->line);
	  if (xref) {
	       sort_refs(s);
	       for (refp = s->xref; refp; refp = refp->next) {
		    if (refp->line != s->line)
			 fprintf(fp, " %4d", refp->line);
	       }
	  }
	  fprintf(fp, "\n");
     }
}

void
tlist(char *fmt, ...)
{
     va_list ap;
     
     if (!lst)
	  return;
     va_start(ap, fmt);
     vfprintf(lst, fmt, ap);
     va_end(ap);
}

/* dump literals and fix forward refs and undefined symbols */
void
symbols()
{
     Symbol *s;
     ref *p;
     
     s = symlist;
     while (s) {
	  if (s->name[0] == '=') {
	       memory[location_counter] = w_eval(0,s->value,5);
	       s->value = location_counter++;
	  }
	  if (s->type == UNDEF
	      && (strlen(s->name) != 2
		  || s->name[1] != 'F'
		  || s->name[0] < '0'
		  || s->name[0] > '9')) {
	       fprintf(stderr,"Warning: '%-10s CON  0' inserted at %d:\n",
		       s->name,location_counter);
	       memory[location_counter] = 0;
	       s->value = location_counter++;
	       s->type = DEF;
	  }
	  s = s->next;
     }
     s = symlist;
     while (s) {
	  /* Now that the literal pool has been dumped we can
	     evaluate the location of the END statement */          
	  if (s->type == END)
	       s->value = location_counter; 
	  /* Now we can handle any forward references */
	  p = s->reflist;
	  while (p) {
	       memory[p->address] &= (B3 - 1);
	       memory[p->address] |= s->value<<(BYTESIZE*3);
	       p = p->next;
	  }
	  if (strlen(s->name) == 2 && s->name[1] == 'F' &&
	      s->name[0] >= '0' && s->name[0] <= '9') {
	       find_forward(s);
	       s->type = DEF;
	  }
	  s = s->next;
     }
}

char loader[] =
" O O6 Z O6    I C O4 0 EH A  F F CF 0  E   EU 0 IH G BB   EJ  CA. Z EU   EH E BA\n"
"   EU 2A-H S BB  C U 1AEH 2AEN V  E  CLU  ABG Z EH E BB J B. A  9               \n";

void
end_routine(int start)
{
     long address, loc, sign, value;
     int i, j, flag; 
     char field[7][11];
     char title[6];
     
     if (source_file) {
	  char *p;
	  if (p = strrchr(source_file, '/'))
	       p++;
	  else
	       p = source_file;
	  for (i = 0; i < 5 && p[i]; i++) 
	       title[i] = toupper(p[i]);
	  title[i] = '\0';
     } else
	  strcpy(title, "MIXAL");
     
     if (writeraw) {
	  int count = 0;
	  for (i = 0; i <= 32; i++) {
	       printf("%c%c%c%c%c",
		      alf[ ((memory[i] & (B5 - 1))>>4*BYTESIZE) & (B1-1) ],
		      alf[ ((memory[i] & (B5 - 1))>>3*BYTESIZE) & (B1-1) ],
		      alf[ ((memory[i] & (B3 - 1))>>2*BYTESIZE) & (B1-1) ],
		      alf[ ((memory[i] & (B2 - 1))>>  BYTESIZE) & (B1-1) ],
		      alf[ (memory[i] & (B1 - 1)) ]);
	       count++;
	       if (count % 16 == 0)
		    printf("\n");
	  }
	  return;
     }
     
     /* The loading routine */
     printf("%s", loader);

     flag = i = 0;
     for (address = 0; address < MEMSIZE; address++) 
	  if (memory[address]) {
	       sign = memory[address] & MINUS ? -1 : 1;
	       value = MAG(memory[address]);
	       if (!flag)
		    flag=1, loc=address;
	       if (address != loc + i || i >= 7) {
		    printf("%-5s%1d%04ld", title, i, loc);
		    for (j = 0; j < i; j++)
			 printf("%10s", field[j]);
		    printf("\n");
		    loc = address;
		    i = 0;
	       } 
	       sprintf(field[i], "%010ld", value);
	       if (sign < 0)
		    field[i][9] = "~JKLMNOPQR"[field[i][9]-'0'];
	       i++;
	  }
     printf("%-5s%1d%04ld", title, i, loc);
     for (j = 0; j < i; j++)
	  printf("%10s", field[j]);
     printf("\n");
     printf("TRANS0%04d\n", start);
}

int
num(char *text) 
{
     int i, n;
     char *p;  

     n = 0;
     for (i = 0; i < 5; i++) 
	  if (!text[i])
	       n = n*B1;
	  else if (p = strchr(alf,text[i]))
	       n = n*B1 + (p - alf);
	  else
	       yyerror("bad character in ALF constant");
     return n;
}

int
eval_op(char *c)
{
     int i;
     for (i = 0; i < op_count; i++)
	  if (!strcmp(c, mix_optab[i].name))
	       return mix_optab[i].c;
     yyerror("invalid opcode.");
     return -1; /* FIXME */
}

int
eval_f(char *c)
{
     int i;
     for (i = 0; i < op_count; i++)
	  if (!strcmp(c, mix_optab[i].name))
	       return mix_optab[i].f;
     return 5;                     
}

int
find_backward(char *name)
{
     char here[3] = "0H";
     int location;  
     int line;
     
     s = symlist;
     here[0] = name[0];
     line = 0;       
     while(s) {
	  if (!strcmp(s->name, here) && lineno > s->line &&
	      s->line > line) {
	       location = s->value;
	       line = s->line;
	  }
	  s = s->next;
     }
     if (line == 0)
	  yyerror("missing backward 'here'");
     return location;
}  

void
find_forward(Symbol *f)
{
     Symbol *s;
     ref *p;
     char here[3] = "0H";
     int location;  
     int line;
     
     s = symlist;
     here[0] = f->name[0];
     p = f->reflist; 
     while (p) {
	  line = B5;               
	  s = symlist;
	  while(s) {
	       if (!strcmp(s->name, here) && p->line < s->line  &&
		   s->line < line) {
		    location = s->value;
		    line = s->line;
	       }
	       s = s->next;
	  }
	  f->value = location;
	  if (line == B5)
	       yyerror("missing 'forward here'");
	  memory[p->address] &= (B3 - 1);
	  memory[p->address] |= location<<(BYTESIZE*3);
	  p = p->next;
     }
}  
